#include <iostream>
#include <stdlib.h>
#include <cmath>
#ifndef AUDIOEDIT_H
#define AUDIOEDIT_H
#include "AudioEdit.h"
#endif
#ifndef FFT_H
#define FFT_H
#include "FFT.h"
#endif

using namespace std;

AudioEdit::AudioEdit()
{
  this->init();
}

AudioEdit::AudioEdit(AudioFile *af)
{
  this->init();
  this->audiofile = af;
  this->setDataLength(this->audiofile->getDataLength());
  this->setDataOriginal(this->audiofile->getDataOriginal());
}

AudioEdit::~AudioEdit()
{
  free((double* )dataStartLength);
  free((double* )dataExpo);
  free((double* )dataFFT);
  free((double* )dataAmpl);
  free((double* )data);
}

void AudioEdit::text(int flag)
{
  switch (flag)
    {
    case 0:
      cout << "#DataLength: " << this->dataLength << endl;
      cout << "#DataOriginal: " << endl;
      for (int i=0;i<this->dataLength;i++)
	{	
	  cout << dec << i << " " << ((double)this->dataOriginal[i]) << endl; 
	}
      break;
    case 1:
      cout << "#DataStartLength: " << endl;
      for (int i=0;i<START_LENGTH;i++)
	{	
	  cout << dec << i << " " << ((double)this->dataStartLength[i]) << endl; 
	}
      break;
    case 2:
      cout << "#DataFFT: " << endl;
      for (int i=0;i</*DOUBLE_LENGTH*/START_LENGTH;i++)
	{	
	  cout << dec << i << " " << ((double)this->dataFFT[i]) << endl; 
	}
      break;
    case 3:
      cout << "#DataAmpl: " << endl;
      for (int i=0;i<HALF_LENGTH;i++)
	{	
	  cout << dec << i << " " << ((double)this->dataAmpl[i]) << endl; 
	}
      break;
    case 4:
      cout << "#Data: " << endl;
      for (int i=0;i<MAX_HAMMING;i++)
	{	
	  cout << dec << i << " " << ((double)this->data[i]) << endl; 
	}
      break;
    default: break;
    }
  cout << endl;
}

void AudioEdit::setDataLength(long length)
{
  this->dataLength = length;
}

void AudioEdit::setDataOriginal(double* data)
{
  this->dataOriginal = data;
  this->orgdataex = true;
}

void AudioEdit::setRedundance(double red)
{
  this->redundance = red;
}

void AudioEdit::setScaleFactor(double sf)
{
  this->scaleFactor = sf;
}

void AudioEdit::setHammingFenster(int anz)
{
  this->hammingFenster = anz;
}

int AudioEdit::getHammingFenster()
{
  return this->hammingFenster;
}

void AudioEdit::init()
{
  this->orgdataex = false;

  this->dataLength = 0;
  this->dataOriginal = NULL;
  this->dataStartLength = (double *)calloc(START_LENGTH, sizeof (double));
  this->dataExpo = (double *)calloc(START_LENGTH, sizeof (double));
  this->dataFFT = (double *)calloc(DOUBLE_LENGTH, sizeof (double));
  this->dataAmpl = (double *)calloc(DOUBLE_LENGTH, sizeof (double)); // sollte hinreichend gross sein
  this->data = (double *)calloc(MAX_HAMMING, sizeof(double));

  this->redundance = 0;
  this->scaleFactor = 0;
  this->hammingFenster = 0;

  this->audiofile = NULL;
}

double* AudioEdit::getDataOrg()
{
  return this->dataOriginal;
}

double* AudioEdit::getDataStartLength()
{
  return this->dataStartLength;
}

double* AudioEdit::getDataFFT()
{
  return this->dataFFT;
}

double* AudioEdit::getData()
{
  return this->data;
}

int AudioEdit::decode()
{
  int b=0;
  if (this->orgdataex == false) return DATAMISSING;

  this->semiinv(); // dataOriginal wird semiinvertiert
  if ((b = this->cutStartLength()) != 0) return b; // aus dataOriginal werden 8K daten ausgeschnitten und in data8192 abgelegt
  this->expoScale(); // der vektor data8192 wird exponentiell skaliert
  this->runFFT(); // fft wird durchgefuehrt
  this->divVector(); // teilt den vector, der die fft hinter sich hat und legt die daten in dataFFT ab -> 8K daten
  this->computeAmpl(); // Amplitudenwerte werden wie folgt berechnet: Ampl = sqrt(Re*Re+Im*Im)
  this->reduceData(); // der datenvektor wird wie gewuenscht reduziert und in data als endergebnis der audiovorverarbeitung abgelegt
  return 0;
}

/* 
 *  absD berechnet den Betrag von double-werten.
 *  parameter: d - wert, von dem der betrag berechnet werden soll
 *  rueckgabe: der betrag von d
 */
double absD(double d)
{
  if (d<0) return (-1)*d;
  else return d;
}

void AudioEdit::reduceData()
{
  double summe_f_hoch_i = 0;
  for (int i=0;i<this->hammingFenster;i++)
    summe_f_hoch_i = summe_f_hoch_i + pow(this->scaleFactor, (double)i);
  int b0 = (int)(HALF_LENGTH/((1-this->redundance)*summe_f_hoch_i));
  if(b0<2) b0=2; /* initiale fensterbreite auf zwei festgelegt */
  int fensterindex = 0;
  for(int i=0;i<this->hammingFenster;i++)
    {
      double mittel = 0;
      fensterindex = fensterindex + (int)(b0*(1-this->redundance));
      b0 = (int)(b0*this->scaleFactor);
      if (b0>HALF_LENGTH) b0=HALF_LENGTH;
     
      for(int j=fensterindex;j<b0+fensterindex;j++)
	mittel = mittel+this->dataAmpl[j];
     
      if (fensterindex+b0>HALF_LENGTH) /* falls fenster ueber vectorgrenze hinauslaeuft */
	  mittel = mittel/(HALF_LENGTH-fensterindex);
      else
	mittel = mittel/b0;
      this->data[i] = mittel;
    }
  /* skalierungsteil */
  double max = data[0];
  double min = data[0];
  for(int i=0;i<this->hammingFenster;i++)
    {
      if(data[i]>max) max=data[i];
      if(data[i]<min) min=data[i];
    }
  double mitte = (max+min)/2;
  for(int i=0;i<this->hammingFenster;i++) {
      data[i] = (data[i]-mitte)/(max-mitte);
      if (isnan(data[i])) {
	cerr << "Ein NaN trat auf und wurde getoetet.";
	data[i] = 0.0;
      }
  }
  // hier data abchecken fuer NaNs falls weitere Probleme auftreten.
	
}

void AudioEdit::computeAmpl()
{
  for (int i=0,j=0;i<HALF_LENGTH;i++,j=j+2)
    this->dataAmpl[i] = sqrt(this->dataFFT[j]*this->dataFFT[j] + this->dataFFT[j+1]*this->dataFFT[j+1]);
}

void AudioEdit::divVector()
{
  double* dummy;
  dummy = (double*)calloc(START_LENGTH, sizeof(double));
  for(int i=0;i<START_LENGTH;i++)
    dummy[i] = this->dataFFT[i];
  free((double*)this->dataFFT);
  this->dataFFT = dummy;
}

void AudioEdit::runFFT()
{
  FFT f;
  for(int i=0,j=0;i<START_LENGTH;i++,j=j+2)
    {
      this->dataFFT[j] = dataExpo[i];
      this->dataFFT[j+1] = 0;
    }
  f.fft(dataFFT-1, START_LENGTH, 1);
}

void AudioEdit::expoScale()
{
  for (int i=0;i<START_LENGTH;i++)
    {
      if (this->dataStartLength[i]<0)
	{
	  this->dataExpo[i] = absD(this->dataStartLength[i]);
	  this->dataExpo[i] = pow(pow(2047, (double)1/127), this->dataExpo[i]);
	  this->dataExpo[i] = this->dataExpo[i] * (-1);
	}
      else
	{
	  this->dataExpo[i] = pow(pow(2047, (double)1/127), this->dataStartLength[i]);
	}
    }
}

int AudioEdit::cutStartLength()
{
  double max = -128;
  int ind = 0;
  long pos = 0;
  double mittel = 0;
  double foranf = 0;

  for (int i=0; i<this->dataLength; i++) // max und mittel ermitteln (es werden nur positive werte betrachtet!)
    {
      if (this->dataOriginal[i]>max) 
	{
	  max = this->dataOriginal[i];
	  ind = i;
	}
      if (this->dataOriginal[i]>0)
	{
	  mittel = mittel + this->dataOriginal[i];
	  pos++;
	}
    }
  mittel = mittel / this->dataLength;
  foranf =  ((max-mittel)/2)+mittel; // foranf stellt den gewuenschten amplitudenwert da (der index ist noch unbekannt) 
  if (max==-128 || foranf<0) return MAXAMPL_ERROR; // anfangsamplitudenwert konnte nicht gefunden werden
  long ind2 = ind;
  while (ind2 > 0) // index fuer anfang ermitteln (im positiven bereich festgelegt!)
    {
      ind2--;
      while(this->dataOriginal[ind2]<=0) // positiven vergleichswert suchen
	{
	  ind2--;
	}
      if (absD(this->dataOriginal[ind2] - foranf) < absD(this->dataOriginal[ind] - foranf)) // anfangsindex bestimmen (welcher index liegt dichter an 0?)
	{
	  ind = ind2;
	}
    }
  if(ind>(this->dataLength-START_LENGTH)) return SIGNALLENGTH_ERROR; // verwendetes signal ist zu kurz!
  for (int i=0; i<START_LENGTH; i++)
    {
      this->dataStartLength[i] = this->dataOriginal[ind]; 
      ind++;
    }
  return 0; // alles gut gegangen
}

void AudioEdit::semiinv()
{
  for(int i=0; i<this->dataLength; i++)
    {
      if ((this->dataOriginal[i]>=0) && (this->dataOriginal[i]<=63))
	{
	  this->dataOriginal[i] = 127 - this->dataOriginal[i];
	}
      else
	{
	  if ((this->dataOriginal[i]>=65) && (this->dataOriginal[i]<=127))
	    {
	      this->dataOriginal[i] = 128 - this->dataOriginal[i];
	    }
	  else
	    {
	      if (this->dataOriginal[i] == 64) this->dataOriginal[i] = 0;
	    }
	}
    }
}
