/* struktur, welche den header einer au-datei darstellt */
struct Header
{
    signed int magic_number; /* hier soll die magicnumber (0x2e736e64 bzw. 0x2e736400) abgelegt werden */
    signed int header_size; /* hier ist die groesse des gesamten headers der au-datei zu finden */
    signed int data_length; /* die datenlaenge des eigentlichen samples soll hier abgelegt werden */
    signed int encoding; /* zeigt an, welche encoding-art vorliegt */
    signed int sample_rate; /* die samplerate des gespeicherten samples wird hier gespeichert */
    signed int nr_of_channels; /* ob mono oder stereo wird hier eingetragen */
};

/* struktur fuer die zusaetzlichen headerbytes */
struct header_add
{
    signed char additional[8]; /* ein array bestehend aus acht elementen, wobei im ersten element der vokal stehen soll (die restlichen elemente werden nicht benutzt) */
};

/* struktur, welche den ausgelesenen datenvektor verwaltet */
struct data
{
    signed char* dataOrg; /* die daten, wie sie in der au-datei stehen */
    double *dataOrgD; /* die gleichen daten, nur als double-werte */
};

/*
 * die klasse AudioFile beinhaltet methoden zum verwalten von au-dateien.
 * zu diesen methoden gehoeren ein konstruktor, der den dateinamen der au-datei uebergeben bekommt, und einen destruktor,
 * sowie die methoden readData, getDataLength, getDataOriginal und text. 
 */
class AudioFile
{
 public:

   /* 
    * der konstruktor ordnet dem objekt den au-dateinamen zu.
    * parameter: dateiname - der dateiname der au-datei
    */ 
  AudioFile(char * dateiname);

   /* destruktor */
  ~AudioFile();

  /* 
   * text gibt daten ueber die au-datei (magicnumber, headersize, etc. okay) - zum debuggen gedacht.
   * parameter: keine, die auszugebenen daten werden den strukturen readHeader bzw. addHeader entnommen
   * rueckgabe: keine, die ausgabe erfolgt auf stdout
   */
  void text();

  /*
   * readData oeffnet eine au-datei und liest sie komplett aus. anschliessend wird verify zum pruefen der daten aufgerufen.
   * parameter: keine, der benoetigte dateiname zum oeffnen der datei wird der membervariable filename entnommen
   * rueckgabe: fehlercode von verify (0 oder -1)
   */
  int readData();
  /* 
   * dieser getter liefert die laenge der originaldaten, welche aus der au-datei ausgelesen wurden.
   * parameter: keine, die laenge wird der struktur readHeader entnommen
   * rueckgabe: die laenge der originaldaten als long-wert
   */
  long getDataLength();

  /*
   * dieser getter liefert einen vektor zurueck, der die originaldaten aus der au-datei als double-werte enthaelt.
   * paramter: keine, der vektor mit den double-werten wird der struktur readDataOrg entnommen.
   * rueckgabe: ein double-pointer auf ein array mit double-werten (enthaelt den originalvektor)
   */
  double* getDataOriginal();
  
  /*
   * create erstellt mit hilfe von audiorecord eine neue au-datei.
   * paramter: vokal - der abzuspeichernde vokal (wird option -i bei audiorecord uebergeben)
   *           length - die aufzunehmende laenge in sekunden (option -t bei audiorecord)
   *           vol - aufnahmelautstaerke (option -v bei audiorecord)
   * rueckgabe: 0
   */
  int create(char vokal, float length, int vol);

  /*
   * play spielt die gerade aufgenommene au-datei mithilfe von audioplay ab.
   * parameter: vol - wiedergabelautstaerke (option -v bei audioplay)
   * rueckgabe: keine
   */
  void play(int vol);

  /*
   * getter, der den vokal (in au-datei abgespeichert) zu pruefzwecken zurueckgibt
   * parameter: keine
   * rueckgabe: ein vokal
   */
  char getVocal();

 private:

  bool mn, hs, dl, en, sr, noc, ahb, zd; /* flags fuer die methode verify (halten fest, ob die einzelnen elemente der au-datei (z.b. magicnumber, headersize, ...) in ordnung sind */
  struct Header readHeader; /* fuer den eingelesenen header */
  struct header_add addHeader; /* fuer die zusaetzlichen headerbytes - hier ein vokal */
  struct data readDataOrg; /* einfache struktur zum verwalten der originaldaten */
  
  /*
   * verify ueberprueft, ob die eingelesenen daten aus der au-datei auch den gewuenschten daten entsprechen (richtiges format, mono, etc.)
   * parameter: keine, daten werden aus den strukturen readHeader und addHeader bezogen
   * rueckgabe: 0 wenn kein fehler aufgetreten ist, sonst -1
   */
  int verify();
  
 protected:							
  char *filename; /* hier ist der dateiname der au-datei gespeichert */
};









