#include "Netz.h"

// Fuer allgemeine Ausgabe und Eingabe
// Format ist jetzt: [schichtenAnzahl]\t[batchAnzahl]\t[trainingsfehlerAnzahl]
//                   [testfehlerAnzahl]\t[epsilon]\t[trainiert]
//                   [Schicht_0]
//                   [Schicht_1] ...
//                   [Schicht_(schichtenAnzahl-1)]
//                   [myStat] (ist mehrere Zeilen)
// mit Eingabe kompatibel!
ostream& operator<< (ostream &ostr, const Netz &theNetz){
  if(((Netz &)theNetz).getGueltigkeit()) {
    ostr << theNetz.schichtenAnzahl << "\t" << theNetz.batchAnzahl
	 << "\t" << theNetz.trainingsfehlerAnzahl << endl;
    ostr << theNetz.testfehlerAnzahl << "\t" << theNetz.epsilon
	 << "\t" << theNetz.trainiert << endl;
    for(int i=0;i<theNetz.schichtenAnzahl;i++) 
      ostr << *(theNetz.Schichte[i]);
    
    ostr << theNetz.myStat;
    ostr << flush;
  }

  return(ostr);
}


// Fuers Testen, und vielleicht auch fuers Laden einer Netzdatei
istream& operator>> (istream &istr, Netz &theNetz){
  theNetz.gueltigkeit = 0;

  istr >> theNetz.schichtenAnzahl;
  istr >> theNetz.batchAnzahl;
  istr >> theNetz.trainingsfehlerAnzahl;
  istr >> theNetz.testfehlerAnzahl;
  istr >> theNetz.epsilon;
  istr >> theNetz.trainiert;

  if (theNetz.schichtenAnzahl > 1) {   // es muss wenigstens 2 Schichte geben
    if(theNetz.Schichte) {
      for(int i=0;i<theNetz.schichtenAnzahl;i++)
	if(theNetz.Schichte[i])
	  delete theNetz.Schichte[i];
      delete[] theNetz.Schichte;   
    }

    theNetz.Schichte = new (Schicht *)[theNetz.schichtenAnzahl];
    if (!theNetz.Schichte) {
      cerr << "Allokationsfehler!" << endl;
      exit(1);
    }
    for(int i=0;i<theNetz.schichtenAnzahl;i++) {
      theNetz.Schichte[i] = new Schicht(2,2,1);    // (2,2,1) sind nur Dummywerte.
      if (!theNetz.Schichte[i]) {
	cerr << "Allokationsfehler!" << endl;
	exit(1);
      }
    }
    for(int i=0;i<theNetz.schichtenAnzahl;i++)
      istr >> *(theNetz.Schichte[i]);              // Dummywerte werden hier ersetzt!

    istr >> theNetz.myStat;
  } else {
    cerr << "Die Schichtenanzahl muss wenigstens 2 sein!\n";
    cerr << "Und sie ist nur " << theNetz.schichtenAnzahl << endl;
    exit(1);
  }

  // wenn es bis hier gelangt ist, dann muesste das Netz eigentlich gueltig sein
  theNetz.gueltigkeit = 1;

  return istr;
}


// schichtenAnzahl, TransferFunktion, neuronenAnzahl pro Schicht, batchAnzahl,
// trainingsfehlerAnzahl, testfehlerAnzahl, epsilon!
// Konstruktor fuer Zufallswerte
Netz::Netz(int theSchichtenAnzahl, int *tf, int *neuronenAnzahl, 
	   int theBatchAnzahl, int theTrainingsfehlerAnzahl, int theTestfehlerAnzahl, 
	   double theEpsilon):schichtenAnzahl(theSchichtenAnzahl),batchAnzahl(theBatchAnzahl),
			      trainingsfehlerAnzahl(theTrainingsfehlerAnzahl),
			      testfehlerAnzahl(theTestfehlerAnzahl),epsilon(theEpsilon),
			      trainiert(0),getTransferFunktionenAusgabe(0),
			      getNeuronenAnzahlAusgabe(0),deltaGewichte(0),
                              deltaSchwellen(0),testCount(0),trainingCount(0),
			      batchCount(0),testPlotRefresh(0),
			      trainingsPlotRefresh(0),gewichtenPlotRefresh(0),
			      testFehler(0),trainingsFehler(0){
  Schichte = 0;

  // Es muss wenigstens 2 Schichte geben, sonst ist es sinnlos!!
  if(schichtenAnzahl > 1) {
    Schichte = new (Schicht *)[schichtenAnzahl];
    if(!Schichte) {
      cerr << "Allokationsfehler!\n";
      exit(1);
    }

    // Die nullte Schicht ist ein Sonderfall (Gewichte sind Identitaetsmatrix):
    // theTf[0] wird immer mit IDENTITAET ersetzt!!!
    // und die Parameter der nullten Schicht werden nie geaendert
    Schichte[0] = new Schicht(neuronenAnzahl[0],neuronenAnzahl[0],IDENTITAET,1);

    for(int i=1;i<schichtenAnzahl;i++){
      Schichte[i] = new Schicht(neuronenAnzahl[i-1],neuronenAnzahl[i],tf[i]);
    }

    // Das Netz muesste eigentlich jetzt funktionieren.
    gueltigkeit = 1;
  }
}



Netz::Netz(void):trainiert(0),Schichte(0),getTransferFunktionenAusgabe(0),getNeuronenAnzahlAusgabe(0),deltaGewichte(0),deltaSchwellen(0),testCount(0),trainingCount(0),batchCount(0),testPlotRefresh(0),trainingsPlotRefresh(0),gewichtenPlotRefresh(0),testFehler(0),trainingsFehler(0){
  gueltigkeit = 0;
}


// Konstruktor, der die Werte von einem fstream & nimmt
Netz::Netz(fstream &source):trainiert(0),Schichte(0),getTransferFunktionenAusgabe(0),
			    getNeuronenAnzahlAusgabe(0),deltaGewichte(0),
                            deltaSchwellen(0),testCount(0),trainingCount(0),
			    batchCount(0),testPlotRefresh(0),
			    trainingsPlotRefresh(0),gewichtenPlotRefresh(0),
			    testFehler(0),trainingsFehler(0){
  source >> *this;
}
  

Netz::~Netz(void){
  if(deltaGewichte){
    for(int i=0;i<schichtenAnzahl;i++) {
      if(deltaGewichte[i]) {
	for(int j=0;j<Schichte[i]->getOutputDim();j++){
	  if(deltaGewichte[i][j])
	    delete[] deltaGewichte[i][j];
	}
	delete[] deltaGewichte[i];
      }
    }
    delete[] deltaGewichte;
  }

  if(deltaSchwellen){
    for(int i=0;i<schichtenAnzahl;i++)
      if(deltaSchwellen[i])
	delete[] deltaSchwellen[i];
    delete[] deltaSchwellen;
  }  

  if(Schichte) {
    for(int i=0;i<schichtenAnzahl;i++)
      if(Schichte[i])
	delete Schichte[i];
    delete[] Schichte;   
  }
}


// ruft doDeltas() von den jeweiligen Schichten auf
void Netz::doDeltas(void){
  if((schichtenAnzahl < 2) || (!Schichte))
    return;
  // tue nicht i=0, weil man die nullte Schicht nicht modifizieren darf!
  for(int i=1;i<schichtenAnzahl;i++)
    Schichte[i]->doDeltas();
  gewichtenPlotRefresh = 1;
}



// Gl. fuer E wie auf S. 18
double Netz::fehlerRechnen(double *sollWerte, double *werte, int laenge){
  double summe = 0.0;
  double temp;

  for(int i=0;i<laenge;i++) {
    temp = sollWerte[i] - werte[i];
    //cout << "E: " << temp << " ";
    summe = summe + temp*temp;
  }

  //cout << ", " << (summe/2.0) << endl;
  return(summe/2.0);
}
  


// propagiert eine Eingabe durch das Netz und gibt
// die Ausgabe zurueck.
double *Netz::propagate(double *input){
  double *temp = input;

  if((schichtenAnzahl < 2) || (!Schichte))
    return 0;
  for(int i=0;i<schichtenAnzahl;i++)
    temp = Schichte[i]->propagate(temp);

  return temp;
}



// Um einmal zu testen!
double *Netz::test(double *input, double *sollOutput){
  // zuerst propagieren
  double *output = propagate(input);

  // Fehler berechnen und speichern in myStat
  testFehler += fehlerRechnen(output, sollOutput,
				   Schichte[schichtenAnzahl-1]->getOutputDim());
  testCount++;
  // Nur manchmal (gemittelten) Fehler speichern:
  if(testCount>=testfehlerAnzahl){
    myStat.addTest(testFehler/(double)testfehlerAnzahl);
    testPlotRefresh = 1;   // muss neu gezeichnet werden!
    testCount = 0;
    testFehler = 0.0;
  }

  return output;
}


// Um einmal zu trainieren!
double *Netz::trainiere(double *input, double *sollOutput){
  if(!deltaGewichte){
    deltaGewichte = new (double **)[schichtenAnzahl];
    if(!deltaGewichte) {
      cerr << "Allokationsfehler!";
      exit(1);
    }
    for(int i=0;i<schichtenAnzahl;i++) {
      deltaGewichte[i] = new (double *)[Schichte[i]->getOutputDim()];
      if(!deltaGewichte[i]) {
	cerr << "Allokationsfehler!";
	exit(1);
      }
      for(int j=0;j<Schichte[i]->getOutputDim();j++){
	deltaGewichte[i][j] = new double[Schichte[i]->getInputDim()];
	if(!deltaGewichte[i][j]) {
	  cerr << "Allokationsfehler!";
	  exit(1);
	}
	for(int k=0;k<Schichte[i]->getInputDim();k++) {
	  deltaGewichte[i][j][k] = 0.0;
	}
      }
    }
  }  // jetzt existiert deltaGewichte auf jeden Fall!

  if(!deltaSchwellen){
    deltaSchwellen  = new (double *)[schichtenAnzahl];
    if(!deltaSchwellen){
      cerr << "Allokationsfehler!";
      exit(1);
    }
    for(int i=0;i<schichtenAnzahl;i++){
      deltaSchwellen[i] = new double[Schichte[i]->getOutputDim()];
      if(!deltaSchwellen[i]){
	cerr << "Allokationsfehler!";
	exit(1);
      }
      for(int j=0;j<Schichte[i]->getOutputDim();j++)
	deltaSchwellen[i][j] = 0.0;
    }
  }  // jetzt existiert deltaSchwellen auf jeden Fall!

  // zuerst propagieren (Gl. 1 und 2):
  double *output = propagate(input);
  
  /*for(int i=0;i<getNeuronenAnzahl()[schichtenAnzahl-1];i++){
    cout << fabs(output[i]-sollOutput[i]) << " ";
  }
  cout << endl;*/
  
  // Gl. 3
  double result;
  if(Schichte[schichtenAnzahl-1]->getTFTyp()==TANH) {
    for(int j=0;j<Schichte[schichtenAnzahl-1]->getOutputDim();j++) {
      result = (-1.0)*(1.0-(output[j]*output[j]))*(sollOutput[j]-output[j]);
      for(int k=0;k<Schichte[schichtenAnzahl-1]->getInputDim();k++)
	deltaGewichte[schichtenAnzahl-1][j][k] = result;
    }
  } else {  // dh, Schichte->getTFTyp()==LOGI
    for(int j=0;j<Schichte[schichtenAnzahl-1]->getOutputDim();j++) {
      result = (-1.0)*(output[j]-(output[j]*output[j]))*(sollOutput[j]-output[j]);
      for(int k=0;k<Schichte[schichtenAnzahl-1]->getInputDim();k++)
	deltaGewichte[schichtenAnzahl-1][j][k] = result;
    }
  }
  
  // Gl. 4
  double summe, temp2;
  for(int i=1;i<schichtenAnzahl-1;i++) {
    for(int j=0;j<Schichte[i]->getOutputDim();j++) {

      temp2 = (Schichte[i]->recentPropagate())[j];
      if(Schichte[i]->getTFTyp()==TANH) {
	temp2 = 1.0 - temp2*temp2;
      }
      else {   // Schichte->getTFTyp()==LOGI
	temp2 = temp2 - temp2*temp2;
      }
      summe = 0.0;
      for(int l=0;l<Schichte[i+1]->getOutputDim();l++)
	summe = summe + deltaGewichte[i+1][l][0]*(Schichte[i+1]->getGewichte())[l][j];
      
      result = summe * temp2;
      for(int k=0;k<Schichte[i]->getInputDim();k++)
	deltaGewichte[i][j][k] = result;
    }
  }

  // Gl. 5 und 6
  double *temp;
  for(int i=1;i<schichtenAnzahl;i++){
    for(int j=0;j<Schichte[i]->getOutputDim();j++){
      deltaSchwellen[i][j] = deltaGewichte[i][j][0] * epsilon; // Gl. 6
      temp = Schichte[i-1]->recentPropagate();
      for(int k=0;k<Schichte[i]->getInputDim();k++)
	deltaGewichte[i][j][k] = (-1.0) * epsilon * deltaGewichte[i][j][k] * temp[k]; // Gl. 5
    }
  }

  
  // die Deltawerte in den Neuronen werden jetzt geupdatet (Gl. 7 und 8)
  for(int i=1;i<schichtenAnzahl;i++){
    Schichte[i]->updateDeltas(deltaGewichte[i], deltaSchwellen[i]);
  }


  // Fuer Online-/Batchlearning (immer noch Gl. 7 und 8)
  batchCount++;
  if(batchCount>=batchAnzahl){
    batchCount = 0;
    //cout << "Updateschritt wird gerade durchgefuehrt!\n";
    doDeltas();
  }


  // Fehler berechnen und speichern in myStat
  trainingsFehler += fehlerRechnen(output, sollOutput,
				   Schichte[schichtenAnzahl-1]->getOutputDim());
  // Nur manchmal den (gemittelten) Fehler speichern
  trainingCount++;
  if(trainingCount>=trainingsfehlerAnzahl){
    //cout << trainingsFehler/(double)trainingsfehlerAnzahl << endl;
    myStat.addTraining(trainingsFehler/(double)trainingsfehlerAnzahl);
    trainingsPlotRefresh = 1;     // muss natuerlich neu gezeichnet werden
    trainingCount = 0;
    trainingsFehler = 0.0;
  }

  return output;
}


GUS &Netz::getGewichtematrix(int schichtIndex){
  if((schichtIndex < schichtenAnzahl) && Schichte){
    myGUS.neuronenAnzahl = Schichte[schichtIndex]->getOutputDim();
    myGUS.inputDim = Schichte[schichtIndex]->getInputDim();
    myGUS.gewichte = Schichte[schichtIndex]->getGewichte();
    myGUS.schwellen = Schichte[schichtIndex]->getSchwellen();
  } 

  return myGUS;
}


int *Netz::getTransferFunktionen(void){
  if (Schichte) {
    if (!getTransferFunktionenAusgabe) {
      getTransferFunktionenAusgabe = new int[schichtenAnzahl];
  
      if (!getTransferFunktionenAusgabe) {
	cerr << "Allokationsfehler!";
	exit(1);
      }
    }

    // Eigentlich haette dieser Teil in der If-Konstruktion auftreten,
    // da die TransferFunktionen nicht geaendert werden duerfen (neues Objekt machen).
    for(int i=0;i<schichtenAnzahl;i++)
      getTransferFunktionenAusgabe[i] = Schichte[i]->getTFTyp();

    return getTransferFunktionenAusgabe;
  } else {
    return 0; // Fehler!
  }
}


int *Netz::getNeuronenAnzahl(void){
  if (Schichte) {
    if(!getNeuronenAnzahlAusgabe){
      getNeuronenAnzahlAusgabe = new int[schichtenAnzahl];
  
      if (!getNeuronenAnzahlAusgabe) {
	cerr << "Allokationsfehler!";
	exit(1);
      }
    }

    // Eigentlich haette dieser Teil in der If-Konstruktion auftreten,
    // da die Neuronenanzahlen nicht geaendert werden duerfen (neues Objekt machen).
    for(int i=0;i<schichtenAnzahl;i++)
      getNeuronenAnzahlAusgabe[i] = Schichte[i]->getOutputDim();

    return getNeuronenAnzahlAusgabe;
  } else {
    return 0; // Fehler!
  }
}


// Fuers Testen und so...
bool Netz::getGueltigkeit(void){
  if (gueltigkeit == 0)
    return gueltigkeit;
  
  if (schichtenAnzahl < 2) {
    cerr << "WARNUNG: (schichtenAnzahl == " << schichtenAnzahl << ") < 2!\n";
    //cerr << "Netz sieht so aus: \n" << *this;
    return 0;
  }
  if (batchAnzahl < 1) {
    cerr << "WARNUNG: (batchAnzahl == " << batchAnzahl << ") < 1!\n";
    //cerr << "Netz sieht so aus: \n" << *this;
    return 0;
  }
  if (trainingsfehlerAnzahl < 1) {
    cerr << "WARNUNG: (trainingsfehlerAnzahl == " << trainingsfehlerAnzahl << ") < 1!\n";
    //cerr << "Netz sieht so aus: \n" << *this;
    return 0;
  }
  if (testfehlerAnzahl < 1) {
    cerr << "WARNUNG: (testfehlerAnzahl == " << testfehlerAnzahl << ") < 1!\n";
    //cerr << "Netz sieht so aus: \n" << *this;
    return 0;
  }
  if ((epsilon < 0.0) || (epsilon > 10.0)) {
    cerr << "WARNUNG: (epsilon == " << epsilon << ") ist nicht im Bereich (0.0 10.0)\n";
    //cerr << "Netz sieht so aus: \n" << *this;
    return 0;
  }
  if (Schichte == 0) {
    cerr << "WARNUNG: Netz besitzt keine Schichten!\n";
    //cerr << "Netz sieht so aus: \n" << *this;
    return 0;
  }
  
  int *temp;
  int i=0;
  temp = getTransferFunktionen();
  if (!temp)
    goto tfProblem;
  for(i=0;i<schichtenAnzahl;i++) {
    if ((temp[i] != TANH) && (temp[i] != LOGI) && (temp[i] != IDENTITAET)) {
    tfProblem:
      cerr << "WARNUNG: Transferfunktionen fuer die " << i << "te Schicht ist nicht im\nBereich TANH, LOGI, IDENTITAET\n";
      //cerr << "Netz sieht so aus: \n" << *this;
      return 0;
    }
  }
   
  temp = getNeuronenAnzahl();
  if (!temp)
    goto nProblem;
  for(i=0;i<schichtenAnzahl;i++) {
    if (temp[i] < 1) {
    nProblem:
      cerr << "WARNUNG: Die " << i << "te Schicht besitzt nur " << temp[i] << " Neuronen!\n";
      //cerr << "Netz sieht so aus: \n" << *this;
      return 0;
    }
  }

  int anzahl = Schichte[0]->getOutputDim();
  for(i=1;i<schichtenAnzahl;i++) {
    if (anzahl != Schichte[i]->getInputDim()) {
      cerr << "WARNUNG: " << (i-1) << "te Schicht hat OutputDim " << anzahl << ", aber " << i << "te Schicht hat InputDim " << Schichte[i]->getInputDim() << "\n";
      //cerr << "Netz sieht so aus: \n" << *this;
      return 0;
    }
    anzahl = Schichte[i]->getOutputDim();
  }
      
  return gueltigkeit;
}

