#ifndef THENETZ
#define THENETZ
 
#include "Schicht.h"
#include <fstream>
#include "Statistik.h"

#include "GUS.h"

class Netz;

ostream& operator<< (ostream &ostr, const Netz &);
istream& operator>> (istream &istr, Netz &); 

class Netz {
  friend ostream& operator<< (ostream &ostr, const Netz &);
  friend istream& operator>> (istream &istr, Netz &);

 protected:
  int schichtenAnzahl;
  int batchAnzahl;  // braucht man dies eigentlich?
  int trainingsfehlerAnzahl;
  int testfehlerAnzahl;
  double epsilon;
  bool trainiert, gueltigkeit;
  Schicht **Schichte;
  Statistik myStat;
  GUS myGUS;
 private:
  // Werden nur fuer Ausgabe benutzt. (Memory-leaks vermeiden)
  // Fuer die jeweiligen Funktionen darf der Benutzer darf sie nicht veraendern.
  int *getTransferFunktionenAusgabe;
  int *getNeuronenAnzahlAusgabe;

  double ***deltaGewichte, **deltaSchwellen;
 protected:
  int testCount;
  int trainingCount;
  int batchCount;
  bool testPlotRefresh;
  bool trainingsPlotRefresh;
  bool gewichtenPlotRefresh;

  double testFehler, trainingsFehler;

 public:
  // schichtenAnzahl, TransferFunktion, neuronenAnzahl pro Schicht, batchAnzahl,
  // trainingsfehlerAnzahl, testfehlerAnzahl, epsilon!
  // Konstruktor fuer Zufallswerte
  Netz(int, int *, int *, int, int, int, double);

  // Damit zuerst ein Objekt haben kann, und nur spaeter durch >> einstellen.
  Netz(void);

  Netz(fstream &source);
  ~Netz(void);

 protected:
  void doDeltas(void);
  static double fehlerRechnen(double *, double *, int);
 public:
  // Eingabe wird durchs Netz propagiert, aber nichts mit Statistik wird gemacht
  double *propagate(double *);
  // Propagation mit Teststatistik
  double *test(double *, double *);
  // Propagation mit Trainingsstatistik und Aenderung der Gewichten/Schwellen
  double *trainiere(double *, double *);

  // Setter und Getter
  GUS &getGewichtematrix(int schichtIndex);
  int *getTransferFunktionen(void);
  int *getNeuronenAnzahl(void);
  bool getGueltigkeit(void);
  
  // inline Setter und Getter
  int getSchichtenAnzahl(void){ return schichtenAnzahl; }
  Statistik &getStatistik(void){ return myStat; }  
  bool getTrainiert(void){ return trainiert; }
  void setTrainiert(bool theTrainiert){ trainiert = theTrainiert; }
  int getBatchAnzahl(void){ return batchAnzahl; }
  void setBatchAnzahl(int theBatchAnzahl){ batchAnzahl = theBatchAnzahl; }
  int getTrainingsfehlerAnzahl(void){ return trainingsfehlerAnzahl; }
  void setTrainingsfehlerAnzahl(int theTrainingsfehlerAnzahl){
    trainingsfehlerAnzahl = theTrainingsfehlerAnzahl; }
  int getTestfehlerAnzahl(void){ return testfehlerAnzahl; }
  void setTestfehlerAnzahl(int theTestfehlerAnzahl){
    testfehlerAnzahl = theTestfehlerAnzahl; }
  double getEpsilon(void){ return epsilon; }
  void setEpsilon(double theEpsilon){ epsilon = theEpsilon; }

  // Jedes Neuzeichnen wird nur ein mal ausgegeben!
  int getTestCount(void){ return testCount; }
  bool getTestPlotRefresh(void){
    bool tmp = testPlotRefresh;
    testPlotRefresh = 0;
    //cout << "testCountZuNull: " << tmp << endl;
    return tmp;
  }
  bool getTrainingsPlotRefresh(void){
    bool tmp = trainingsPlotRefresh;
    trainingsPlotRefresh = 0;
    //cout << "trainingsPlotRefresh: " << tmp << endl;
    return tmp;
  }
  bool getGewichtenPlotRefresh(void){
    bool tmp = gewichtenPlotRefresh;
    gewichtenPlotRefresh = 0;
    //cout << "gewichtenPlotRefresh: " << tmp << endl << flush;
    return tmp;
  }
};

#endif
